@@ -251,7 +251,8 @@ static int csv_config (const char *key, const char *value)
 	return (0);
 } /* int csv_config */
 
-static int csv_write (const data_set_t *ds, const value_list_t *vl)
+static int csv_write (const data_set_t *ds, const value_list_t *vl,
+		user_data_t __attribute__((unused)) *user_data)
 {
 	struct stat  statbuf;
 	char         filename[512];
@@ -358,5 +359,5 @@ void module_register (void)
 {
 	plugin_register_config ("csv", csv_config,
 			config_keys, config_keys_num);
-	plugin_register_write ("csv", csv_write);
+	plugin_register_write ("csv", csv_write, /* user_data = */ NULL);
 } /* void module_register */
