@@ -314,7 +314,8 @@ static int rc_init (void)
   return (0);
 } /* int rc_init */
 
-static int rc_write (const data_set_t *ds, const value_list_t *vl)
+static int rc_write (const data_set_t *ds, const value_list_t *vl,
+    user_data_t __attribute__((unused)) *user_data)
 {
   char filename[512];
   char values[512];
@@ -405,7 +406,7 @@ void module_register (void)
   plugin_register_config ("rrdcached", rc_config,
       config_keys, config_keys_num);
   plugin_register_init ("rrdcached", rc_init);
-  plugin_register_write ("rrdcached", rc_write);
+  plugin_register_write ("rrdcached", rc_write, /* user_data = */ NULL);
   plugin_register_shutdown ("rrdcached", rc_shutdown);
 } /* void module_register */
 
